import os
import tempfile
import pytest
import numpy as np
from unittest.mock import patch, MagicMock
import yaml

from napari_simpleannotate import BboxVideoQWidget


def test_bbox_video_widget_init(make_napari_viewer):
    """Test BboxVideoQWidget initialization."""
    viewer = make_napari_viewer()
    widget = BboxVideoQWidget(viewer)
    
    # Check that widget is properly initialized
    assert widget.viewer == viewer
    assert hasattr(widget, 'classlistWidget')
    assert hasattr(widget, 'class_textbox')
    assert hasattr(widget, 'video_info_label')
    assert hasattr(widget, 'frame_info_label')
    assert hasattr(widget, 'features')
    assert widget.video_path == ""
    assert widget.video_dir == ""
    assert widget.total_frames == 0
    assert widget.current_frame == 0


def test_add_class(make_napari_viewer):
    """Test adding a class to the class list."""
    viewer = make_napari_viewer()
    widget = BboxVideoQWidget(viewer)
    
    with tempfile.TemporaryDirectory() as temp_dir:
        widget.video_dir = temp_dir
        
        # Add a class
        widget.class_textbox.setText("person")
        
        with patch.object(widget, 'save_classes') as mock_save:
            widget.add_class()
        
        # Check that class was added
        assert widget.classlistWidget.count() == 1
        assert widget.classlistWidget.item(0).text() == "0: person"
        assert widget.class_textbox.text() == ""  # Should be cleared
        mock_save.assert_called_once()


def test_add_duplicate_class(make_napari_viewer):
    """Test that duplicate classes show info message."""
    viewer = make_napari_viewer()
    widget = BboxVideoQWidget(viewer)
    
    # Add a class first
    widget.class_textbox.setText("person")
    widget.add_class()
    
    # Try to add the same class again
    widget.class_textbox.setText("person")
    
    with patch('qtpy.QtWidgets.QMessageBox.information') as mock_info:
        widget.add_class()
        mock_info.assert_called_once()
    
    # Should still have only one class
    assert widget.classlistWidget.count() == 1


def test_del_class(make_napari_viewer):
    """Test deleting a class from the class list."""
    viewer = make_napari_viewer()
    widget = BboxVideoQWidget(viewer)
    
    with tempfile.TemporaryDirectory() as temp_dir:
        widget.video_dir = temp_dir
        
        # Add a class
        widget.classlistWidget.addItem("0: person")
        widget.classlistWidget.setCurrentRow(0)
        
        with patch('qtpy.QtWidgets.QMessageBox.question', return_value=16384):  # QMessageBox.Yes
            with patch.object(widget, 'save_classes') as mock_save:
                widget.del_class()
                mock_save.assert_called_once()
        
        # Check that class was deleted
        assert widget.classlistWidget.count() == 0


def test_sort_classlist(make_napari_viewer):
    """Test that class list is properly sorted."""
    viewer = make_napari_viewer()
    widget = BboxVideoQWidget(viewer)
    
    # Add classes in non-sequential order
    widget.classlistWidget.addItem("2: car")
    widget.classlistWidget.addItem("0: person")
    widget.classlistWidget.addItem("1: bike")
    
    widget.sort_classlist()
    
    # Check order
    assert widget.classlistWidget.item(0).text() == "0: person"
    assert widget.classlistWidget.item(1).text() == "1: bike"
    assert widget.classlistWidget.item(2).text() == "2: car"


def test_load_classes(make_napari_viewer):
    """Test loading classes from class.yaml file."""
    viewer = make_napari_viewer()
    widget = BboxVideoQWidget(viewer)
    
    with tempfile.TemporaryDirectory() as temp_dir:
        widget.video_dir = temp_dir
        
        # Create class.yaml file
        class_data = {0: "person", 1: "car", 2: "bike"}
        class_file = os.path.join(temp_dir, "class.yaml")
        with open(class_file, 'w') as f:
            yaml.dump(class_data, f)
        
        widget.load_classes()
        
        # Check that classes were loaded and sorted
        assert widget.classlistWidget.count() == 3
        assert widget.classlistWidget.item(0).text() == "0: person"
        assert widget.classlistWidget.item(1).text() == "1: car"
        assert widget.classlistWidget.item(2).text() == "2: bike"


def test_save_classes(make_napari_viewer):
    """Test saving classes to class.yaml file."""
    viewer = make_napari_viewer()
    widget = BboxVideoQWidget(viewer)
    
    with tempfile.TemporaryDirectory() as temp_dir:
        widget.video_dir = temp_dir
        
        # Add some classes
        widget.classlistWidget.addItem("0: person")
        widget.classlistWidget.addItem("1: car")
        
        widget.save_classes()
        
        # Check that class.yaml file was created
        class_file = os.path.join(temp_dir, "class.yaml")
        assert os.path.exists(class_file)
        
        # Check file contents
        with open(class_file, 'r') as f:
            loaded_data = yaml.safe_load(f)
        
        assert loaded_data[0] == "person"
        assert loaded_data[1] == "car"


@patch('qtpy.QtWidgets.QFileDialog.getOpenFileName')
@patch('napari_simpleannotate._bbox_video_widget.HAS_PYAV', True)
def test_open_video_no_file_selected(mock_dialog, make_napari_viewer):
    """Test opening video when no file is selected."""
    viewer = make_napari_viewer()
    widget = BboxVideoQWidget(viewer)
    
    # Mock file dialog returning no file
    mock_dialog.return_value = ("", "")
    
    widget.openVideo()
    
    # Should not crash and video_path should remain empty
    assert widget.video_path == ""


@patch('napari_simpleannotate._bbox_video_widget.HAS_PYAV', False)
def test_open_video_no_pyav(make_napari_viewer):
    """Test opening video when PyAV is not available."""
    viewer = make_napari_viewer()
    widget = BboxVideoQWidget(viewer)
    
    with patch('qtpy.QtWidgets.QMessageBox.warning') as mock_warning:
        widget.openVideo()
        mock_warning.assert_called_once()


def test_on_frame_changed(make_napari_viewer):
    """Test frame change event handling."""
    viewer = make_napari_viewer()
    widget = BboxVideoQWidget(viewer)
    
    # Mock event with frame value
    mock_event = MagicMock()
    mock_event.value = [5, 0, 0]  # frame 5
    
    with patch.object(widget, 'update_frame_info') as mock_update:
        widget.on_frame_changed(mock_event)
        
        assert widget.current_frame == 5
        mock_update.assert_called_once()


def test_update_frame_info(make_napari_viewer):
    """Test frame information display update."""
    viewer = make_napari_viewer()
    widget = BboxVideoQWidget(viewer)
    
    widget.current_frame = 10
    widget.total_frames = 100
    
    widget.update_frame_info()
    
    assert widget.frame_info_label.text() == "Frame: 10/100"


def test_class_clicked(make_napari_viewer):
    """Test class selection functionality."""
    viewer = make_napari_viewer()
    widget = BboxVideoQWidget(viewer)
    
    # Add a class and select it
    widget.classlistWidget.addItem("0: person")
    widget.classlistWidget.setCurrentRow(0)
    widget.current_frame = 5
    
    # Mock shapes layer
    bbox_layer = viewer.layers["bbox_layer"]
    bbox_layer.selected_data = {0}  # One shape selected
    bbox_layer.features = {"class": ["old_class"], "frame": [0]}
    
    widget.class_clicked()
    
    # Check that feature defaults were updated
    assert bbox_layer.feature_defaults["class"] == "0: person"


def test_layers_initialization(make_napari_viewer):
    """Test that napari layers are properly initialized."""
    viewer = make_napari_viewer()
    widget = BboxVideoQWidget(viewer)
    
    # Check that bbox layer was added
    layer_names = [layer.name for layer in viewer.layers]
    assert "bbox_layer" in layer_names
    
    # Check bbox layer properties
    bbox_layer = viewer.layers["bbox_layer"]
    assert bbox_layer.features == widget.features
    assert bbox_layer.text == widget.text


@patch('qtpy.QtWidgets.QFileDialog.getOpenFileName')
@patch('napari_simpleannotate._bbox_video_widget.HAS_PYAV', True)
def test_load_video_success(mock_dialog, make_napari_viewer):
    """Test successful video loading."""
    viewer = make_napari_viewer()
    widget = BboxVideoQWidget(viewer)
    
    # Setup mocks
    test_video_path = "/path/to/test.mp4"
    mock_dialog.return_value = (test_video_path, "")
    
    # Mock the PyAV loading
    mock_frames = np.random.randint(0, 255, (100, 480, 640, 3), dtype=np.uint8)
    
    with patch.object(widget, 'load_video_with_pyav', return_value=mock_frames) as mock_load_pyav:
        with patch.object(widget, 'load_classes') as mock_load_classes:
            with patch.object(widget, 'load_annotations') as mock_load_annotations:
                widget.openVideo()
    
    # Check that video was loaded
    assert widget.video_path == test_video_path
    assert widget.video_dir == "/path/to"
    assert widget.total_frames == 100
    assert "test.mp4" in widget.video_info_label.text()
    mock_load_pyav.assert_called_once_with(test_video_path)
    mock_load_classes.assert_called_once()
    mock_load_annotations.assert_called_once()